#!/usr/bin/env python3
"""
Virus Data Retrieval Module for Report 001

This module retrieves virus metagenomic data from LabKey using executeSQL API
based on the virus.toml configuration. It returns a pandas DataFrame containing
only the essential fields used by labkey_files/names.sql for efficient processing.

Usage as a module:
    from retrieve_virus_data import retrieve_all_virus_data, retrieve_virus_data_by_visibility
    
    # Get all virus data
    df = retrieve_all_virus_data()
    
    # Get only public viruses
    public_df = retrieve_virus_data_by_visibility('public')
"""

import sys
import os
from pathlib import Path
from typing import Dict, List, Any, Optional, Union
import pandas as pd
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
import pickle
import gzip
from datetime import datetime

# Load environment variables from .env file if running locally
try:
    from dotenv import load_dotenv
    # Look for .env file in the project root
    env_path = Path(__file__).parent.parent.parent / '.env'
    if env_path.exists():
        load_dotenv(env_path)
except ImportError:
    # dotenv not available, rely on environment variables being set externally
    pass

# Add the labkey_api module to the path
sys.path.insert(0, str(Path(__file__).parent.parent.parent / 'labkey_api'))

from utils.labkey_client import LabKeyClient
from virus_query_generator import VirusQueryGenerator


def apply_lower_to_sql_expression(expression: str) -> str:
    """
    Apply LOWER() function to field names in a SQL expression while preserving structure.
    
    Handles complex expressions with OR, AND, NOT, and parentheses.
    
    Args:
        expression: SQL expression to process
        
    Returns:
        Expression with LOWER() applied to field names
    """
    import re
    
    # Map of lowercase field names to proper LabKey schema names
    field_mapping = {
        'rank': 'Rank',
        'species': 'Species', 
        'stitle': 'Stitle'
    }
    
    # Pattern to match field names followed by LIKE or NOT LIKE
    # This regex captures: (field name) (optional NOT) (LIKE) (pattern)
    pattern = r'\b(rank|species|stitle)\s+(NOT\s+)?LIKE\s+'
    
    def replace_field(match):
        field = match.group(1)
        not_part = match.group(2) or ''  # Will be 'NOT ' or empty
        
        # Map to proper case
        proper_field = field_mapping.get(field.lower(), field)
        
        # Apply LOWER to the field
        return f"LOWER({proper_field}) {not_part}LIKE "
    
    # Apply the replacement
    result = re.sub(pattern, replace_field, expression, flags=re.IGNORECASE)
    
    # Handle patterns - lowercase the content inside quotes
    pattern_regex = r"LIKE\s+'([^']+)'"
    def lowercase_pattern(match):
        content = match.group(1).lower()
        return f"LIKE '{content}'"
    
    result = re.sub(pattern_regex, lowercase_pattern, result)
    
    return result


def get_essential_fields() -> List[str]:
    """
    Get the essential fields needed based on labkey_files/names.sql.
    
    These are the minimum fields required for virus processing pipeline.
    Field names match the actual LabKey schema (capitalized).
    
    Returns:
        List of field names to retrieve from LabKey
    """
    return [
        'sewershed_location',  # This one is lowercase
        'sample_date',         # This one is lowercase
        'Rank',               # Capitalized in LabKey
        'Species',            # Capitalized in LabKey
        'Stitle',             # Capitalized in LabKey
        'Mapped_Reads',       # Capitalized in LabKey
        'Total_Reads'         # Capitalized in LabKey
    ]


def create_optimized_virus_query(virus_generator: VirusQueryGenerator, 
                                virus_keys: List[str],
                                max_rows: Optional[int] = None,
                                test_filter: bool = False) -> str:
    """
    Create an optimized SQL query to retrieve virus data with only essential fields.
    
    Args:
        virus_generator: VirusQueryGenerator instance
        virus_keys: List of virus keys to include in query
        max_rows: Maximum number of rows to return (None for unlimited)
        test_filter: If True, add WHERE clause to filter Columbia, MO June 2024 for fast testing
        
    Returns:
        Optimized SQL query string
    """
    if not virus_keys:
        raise ValueError("At least one virus key must be provided")
    
    config = virus_generator.config
    query_settings = config.get('query_settings', {})
    base_table = query_settings.get('base_table', 'metagenomic_hits_top_hit_etl')
    required_filters = query_settings.get('required_filters', [])
    
    # Use only essential fields for efficiency
    essential_fields = get_essential_fields()
    fields_str = ', '.join(essential_fields)
    
    # Build virus detection conditions
    virus_conditions = []
    for virus_key in virus_keys:
        virus_config = config.get('virus', {}).get(virus_key)
        if not virus_config:
            continue
            
        search_criteria = virus_config.get('search_criteria', [])
        if not search_criteria:
            continue
        
        if len(search_criteria) == 1:
            # Single criterion - might contain OR/AND internally
            condition = apply_lower_to_sql_expression(search_criteria[0])
        else:
            # Multiple criteria - need to AND them together
            fixed_conditions = []
            for criterion in search_criteria:
                fixed_condition = apply_lower_to_sql_expression(criterion)
                fixed_conditions.append(fixed_condition)
            condition = f"({' AND '.join(fixed_conditions)})"
        
        virus_conditions.append(condition)
    
    # Combine all virus conditions with OR
    if not virus_conditions:
        raise ValueError("No valid virus conditions found - check virus.toml configuration")
    
    virus_filter = f"({' OR '.join(virus_conditions)})"
    
    # Build complete WHERE clause
    all_filters = required_filters + [virus_filter]
    
    # Add test filter for Columbia, MO June 2024 if requested
    if test_filter:
        test_conditions = [
            "sewershed_location = 'Columbia, MO'",
            "sample_date >= '2024-06-01'",
            "sample_date < '2024-07-01'"
        ]
        all_filters.extend(test_conditions)
    
    where_clause = ' AND '.join(all_filters)
    
    # Build final query (without line breaks to avoid SQL parsing issues)
    sql = f"SELECT {fields_str} FROM {base_table} WHERE {where_clause}"
    
    if max_rows:
        sql += f" LIMIT {max_rows}"
    
    return sql.strip()



def retrieve_all_virus_data_from_config(max_rows: Optional[int] = None,
                                       batch_size: int = 50000,
                                       show_progress: bool = True,
                                       test_filter: bool = False) -> pd.DataFrame:
    """
    Retrieve all virus data specified in virus.toml configuration.
    
    This function retrieves all viruses regardless of their public/private designation,
    as the visibility classification will be applied later in the processing pipeline.
    
    Args:
        max_rows: Maximum total rows to return (None for unlimited)
        batch_size: Rows per batch for large datasets (not used after removing batching)
        show_progress: Whether to show progress indicators
        test_filter: If True, add WHERE clause to filter Columbia, MO June 2024 for fast testing
        
    Returns:
        pandas DataFrame with all virus metagenomic data from virus.toml
        
    Raises:
        ValueError: If configuration or connection fails
        Exception: If data retrieval fails
    """
    console = Console()
    
    try:
        # Initialize clients
        base_path = Path(__file__).parent
        labkey_client = LabKeyClient(base_path)
        virus_generator = VirusQueryGenerator()
        
        if show_progress:
            filter_msg = " (Columbia, MO June 2024 only)" if test_filter else ""
            console.print(f"[blue]Retrieving all virus data from LabKey based on virus.toml{filter_msg}...[/blue]")
        
        # Get all virus keys from configuration
        all_viruses = virus_generator.get_private_viruses()  # Returns all viruses
        all_virus_keys = list(all_viruses.keys())
        
        if not all_virus_keys:
            console.print(f"[yellow]No viruses found in virus.toml configuration[/yellow]")
            return pd.DataFrame()
        
        if show_progress:
            console.print(f"[green]Found {len(all_virus_keys)} viruses configured in virus.toml[/green]")
            console.print(f"[cyan]Retrieving all viruses in a single query...[/cyan]")
        
        # Create single SQL query with all viruses
        sql = create_optimized_virus_query(virus_generator, all_virus_keys, max_rows, test_filter)
        
        # Execute single query
        if show_progress:
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                TimeElapsedColumn(),
                console=console,
            ) as progress:
                task = progress.add_task("Executing query...", total=None)
                
                try:
                    result = labkey_client.execute_sql(schema_name="lists", sql=sql)
                    df = pd.DataFrame(result.get('rows', []))
                    progress.update(task, description=f"Retrieved {len(df)} records", completed=100)
                except Exception as e:
                    progress.update(task, description=f"Query failed", completed=100)
                    raise e
        else:
            # Silent retrieval
            result = labkey_client.execute_sql(schema_name="lists", sql=sql)
            df = pd.DataFrame(result.get('rows', []))
        
        # Apply max_rows limit if specified and we have data
        if max_rows and len(df) > max_rows:
            df = df.head(max_rows)
        
        # Basic data cleaning and optimization
        if not df.empty:
            # Convert sample_date to datetime if possible
            if 'sample_date' in df.columns:
                df['sample_date'] = pd.to_datetime(df['sample_date'], errors='coerce')
            
            # Convert numeric columns
            numeric_columns = ['mapped_reads', 'total_reads']
            for col in numeric_columns:
                if col in df.columns:
                    df[col] = pd.to_numeric(df[col], errors='coerce')
            
            # Remove any completely null rows
            df = df.dropna(how='all')
        
        if show_progress:
            console.print(f"[green]✓ Successfully retrieved {len(df)} virus records[/green]")
        
        return df
        
    except Exception as e:
        error_msg = f"Failed to retrieve all virus data: {e}"
        if show_progress:
            console.print(f"[red]✗ {error_msg}[/red]")
        raise Exception(error_msg)


def save_to_cache(df: pd.DataFrame, cache_file: Path, metadata: Dict[str, Any]) -> None:
    """Save DataFrame to compressed pickle with metadata."""
    cache_data = {
        'data': df,
        'metadata': metadata,
        'timestamp': datetime.now(),
        'version': '1.0'
    }
    
    cache_file.parent.mkdir(parents=True, exist_ok=True)
    
    with gzip.open(cache_file, 'wb') as f:
        pickle.dump(cache_data, f, protocol=pickle.HIGHEST_PROTOCOL)


def load_from_cache(cache_file: Path) -> tuple[pd.DataFrame, Dict[str, Any]]:
    """Load DataFrame from compressed pickle with metadata."""
    with gzip.open(cache_file, 'rb') as f:
        cache_data = pickle.load(f)
    
    return cache_data['data'], cache_data


# Alias for backward compatibility and clearer naming
def retrieve_all_virus_data(max_rows: Optional[int] = None,
                           batch_size: int = 50000,
                           show_progress: bool = True,
                           test_filter: bool = False,
                           cache_file: Optional[Path] = None,
                           use_cache: bool = False) -> pd.DataFrame:
    """
    Retrieve all virus data from virus.toml configuration.
    
    This is the main function to use for retrieving virus data.
    It gets all viruses specified in virus.toml regardless of visibility.
    
    Args:
        max_rows: Maximum total rows to return (None for unlimited)
        batch_size: Rows per batch for large datasets  
        show_progress: Whether to show progress indicators
        test_filter: If True, filter Columbia, MO June 2024 for fast testing
        cache_file: Path to cache file for saving/loading data
        use_cache: If True and cache exists, load from cache instead of querying
        
    Returns:
        pandas DataFrame with all virus metagenomic data
    """
    console = Console()
    
    # Try to load from cache if requested
    if use_cache and cache_file and cache_file.exists():
        if show_progress:
            console.print(f"[cyan]Loading virus data from cache: {cache_file}[/cyan]")
        
        try:
            df, cache_info = load_from_cache(cache_file)
            age_days = (datetime.now() - cache_info['timestamp']).days
            
            if show_progress:
                console.print(f"[green]✓ Loaded {len(df)} records from cache[/green]")
                console.print(f"  Cache age: {age_days} days old")
                console.print(f"  Created: {cache_info['timestamp'].strftime('%Y-%m-%d %H:%M:%S')}")
                
                if age_days > 7:
                    console.print(f"[yellow]⚠ Cache is {age_days} days old. Consider refreshing.[/yellow]")
            
            return df
            
        except Exception as e:
            console.print(f"[yellow]Failed to load cache: {e}[/yellow]")
            console.print("[cyan]Falling back to fresh data retrieval...[/cyan]")
    
    # Retrieve fresh data
    df = retrieve_all_virus_data_from_config(
        max_rows=max_rows,
        batch_size=batch_size,
        show_progress=show_progress,
        test_filter=test_filter
    )
    
    # Save to cache if cache_file is specified
    if cache_file and not df.empty:
        if show_progress:
            console.print(f"[cyan]Saving data to cache: {cache_file}[/cyan]")
        
        # Create enhanced metadata for freshness checking
        metadata = {
            'max_rows': max_rows,
            'test_filter': test_filter,
            'row_count': len(df),
            'columns': list(df.columns),
            'total_records': len(df)
        }
        
        # Add sample date information if available
        if not df.empty and 'sample_date' in df.columns:
            metadata['latest_sample_date'] = df['sample_date'].max()
            metadata['earliest_sample_date'] = df['sample_date'].min()
            metadata['unique_locations'] = df['sewershed_location'].nunique() if 'sewershed_location' in df.columns else 0
            metadata['total_mapped_reads'] = df['mapped_reads'].sum() if 'mapped_reads' in df.columns else 0
        
        try:
            save_to_cache(df, cache_file, metadata)
            if show_progress:
                console.print(f"[green]✓ Saved {len(df)} records to cache[/green]")
        except Exception as e:
            console.print(f"[yellow]Failed to save cache: {e}[/yellow]")
    
    return df


def retrieve_virus_data_by_visibility(visibility: str = 'public',
                                    max_rows: Optional[int] = None,
                                    batch_size: int = 50000,
                                    show_progress: bool = True) -> pd.DataFrame:
    """
    Retrieve virus data filtered by visibility level.
    
    NOTE: This function is kept for backward compatibility, but the recommended
    approach is to use retrieve_all_virus_data() and apply visibility filtering
    later in the processing pipeline.
    
    Args:
        visibility: 'public' or 'private' (private includes all viruses)
        max_rows: Maximum total rows to return (None for unlimited)
        batch_size: Rows per batch for large datasets
        show_progress: Whether to show progress indicators
        
    Returns:
        pandas DataFrame with virus metagenomic data
    """
    console = Console()
    
    try:
        # Initialize clients
        base_path = Path(__file__).parent
        labkey_client = LabKeyClient(base_path)
        virus_generator = VirusQueryGenerator()
        
        if show_progress:
            console.print(f"[blue]Retrieving {visibility} virus data from LabKey...[/blue]")
        
        # Get virus keys by visibility
        if visibility == 'public':
            viruses = virus_generator.get_public_viruses()
        elif visibility == 'private':
            viruses = virus_generator.get_private_viruses()  # Returns all viruses
        else:
            raise ValueError(f"Invalid visibility: {visibility}. Must be 'public' or 'private'")
        
        virus_keys = list(viruses.keys())
        
        if not virus_keys:
            console.print(f"[yellow]No {visibility} viruses found in configuration[/yellow]")
            return pd.DataFrame()
        
        if show_progress:
            console.print(f"[green]Found {len(virus_keys)} {visibility} viruses to retrieve[/green]")
        
        # Create single SQL query with all viruses
        sql = create_optimized_virus_query(virus_generator, virus_keys, max_rows, test_filter=False)
        
        # Execute single query
        if show_progress:
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                TimeElapsedColumn(),
                console=console,
            ) as progress:
                task = progress.add_task("Executing query...", total=None)
                
                try:
                    result = labkey_client.execute_sql(schema_name="lists", sql=sql)
                    df = pd.DataFrame(result.get('rows', []))
                    progress.update(task, description=f"Retrieved {len(df)} records", completed=100)
                except Exception as e:
                    progress.update(task, description=f"Query failed", completed=100)
                    raise e
        else:
            # Silent retrieval
            result = labkey_client.execute_sql(schema_name="lists", sql=sql)
            df = pd.DataFrame(result.get('rows', []))
        
        # Apply max_rows limit if specified and we have data
        if max_rows and len(df) > max_rows:
            df = df.head(max_rows)
        
        # Basic data cleaning and optimization
        if not df.empty:
            # Convert sample_date to datetime if possible
            if 'sample_date' in df.columns:
                df['sample_date'] = pd.to_datetime(df['sample_date'], errors='coerce')
            
            # Convert numeric columns
            numeric_columns = ['mapped_reads', 'total_reads']
            for col in numeric_columns:
                if col in df.columns:
                    df[col] = pd.to_numeric(df[col], errors='coerce')
            
            # Remove any completely null rows
            df = df.dropna(how='all')
        
        if show_progress:
            console.print(f"[green]✓ Successfully retrieved {len(df)} virus records[/green]")
        
        return df
        
    except Exception as e:
        error_msg = f"Failed to retrieve {visibility} virus data: {e}"
        if show_progress:
            console.print(f"[red]✗ {error_msg}[/red]")
        raise Exception(error_msg)


def retrieve_virus_data_by_type(virus_type: str,
                               max_rows: Optional[int] = None,
                               batch_size: int = 50000,
                               show_progress: bool = True) -> pd.DataFrame:
    """
    Retrieve virus data filtered by virus type.
    
    Args:
        virus_type: 'enteric', 'respiratory', 'atypical', or 'animal'
        max_rows: Maximum total rows to return (None for unlimited)
        batch_size: Rows per batch for large datasets
        show_progress: Whether to show progress indicators
        
    Returns:
        pandas DataFrame with virus metagenomic data for specified type
    """
    console = Console()
    
    try:
        # Initialize clients
        base_path = Path(__file__).parent
        labkey_client = LabKeyClient(base_path)
        virus_generator = VirusQueryGenerator()
        
        if show_progress:
            console.print(f"[blue]Retrieving {virus_type} virus data from LabKey...[/blue]")
        
        # Get virus keys by type
        viruses = virus_generator.get_viruses_by_type(virus_type)
        virus_keys = list(viruses.keys())
        
        if not virus_keys:
            console.print(f"[yellow]No {virus_type} viruses found in configuration[/yellow]")
            return pd.DataFrame()
        
        if show_progress:
            console.print(f"[green]Found {len(virus_keys)} {virus_type} viruses to retrieve[/green]")
        
        # Create single SQL query with all viruses of this type
        sql = create_optimized_virus_query(virus_generator, virus_keys, max_rows, test_filter=False)
        
        # Execute single query
        result = labkey_client.execute_sql(schema_name="lists", sql=sql)
        df = pd.DataFrame(result.get('rows', []))
        
        # Apply max_rows limit if specified
        if max_rows and len(df) > max_rows:
            df = df.head(max_rows)
        
        if show_progress:
            console.print(f"[green]✓ Successfully retrieved {len(df)} {virus_type} virus records[/green]")
        
        return df
        
    except Exception as e:
        error_msg = f"Failed to retrieve {virus_type} virus data: {e}"
        if show_progress:
            console.print(f"[red]✗ {error_msg}[/red]")
        raise Exception(error_msg)


def get_data_summary(df: pd.DataFrame) -> Dict[str, Any]:
    """
    Get summary statistics for the retrieved virus data.
    
    Args:
        df: pandas DataFrame with virus data
        
    Returns:
        Dictionary with summary statistics
    """
    if df.empty:
        return {"total_records": 0, "message": "No data available"}
    
    summary = {
        "total_records": len(df),
        "date_range": {
            "earliest": df['sample_date'].min() if 'sample_date' in df.columns else None,
            "latest": df['sample_date'].max() if 'sample_date' in df.columns else None
        },
        "unique_locations": df['sewershed_location'].nunique() if 'sewershed_location' in df.columns else 0,
        "total_mapped_reads": df['mapped_reads'].sum() if 'mapped_reads' in df.columns else 0,
        "columns": list(df.columns)
    }
    
    return summary


def main():
    """Demonstration of virus data retrieval functionality."""
    console = Console()
    
    console.print("[bold green]🦠 Virus Data Retrieval Module for Report 001[/bold green]\n")
    
    # Check if API key is available
    if not os.environ.get('LABKEY_API'):
        console.print("[yellow]⚠ LABKEY_API environment variable not set[/yellow]")
        console.print("Set your LabKey API key to test data retrieval:")
        console.print("export LABKEY_API='your_api_key_here'")
        return 1
    
    try:
        # Test 1: Retrieve all virus data (main function)
        console.print("[blue]Test 1: Retrieving all virus data from virus.toml (10 records sample)...[/blue]")
        all_virus_sample = retrieve_all_virus_data(max_rows=10)
        
        console.print(f"✓ Retrieved {len(all_virus_sample)} virus records from virus.toml")
        if not all_virus_sample.empty:
            console.print(f"Columns: {list(all_virus_sample.columns)}")
            console.print(f"Sample locations: {all_virus_sample['sewershed_location'].unique()[:3].tolist()}")
        
        # Test 2: Retrieve enteric viruses (for comparison)
        console.print("\n[blue]Test 2: Retrieving enteric virus data (20 records)...[/blue]")
        enteric_sample = retrieve_virus_data_by_type('enteric', max_rows=20)
        
        console.print(f"✓ Retrieved {len(enteric_sample)} enteric virus records")
        
        # Test 3: Get data summary
        if not all_virus_sample.empty:
            console.print("\n[blue]Test 3: Data summary...[/blue]")
            summary = get_data_summary(all_virus_sample)
            
            for key, value in summary.items():
                if key != "columns":
                    console.print(f"  {key}: {value}")
        
        console.print("\n[green]✅ All tests completed successfully![/green]")
        console.print("\n[cyan]Primary usage as a module:[/cyan]")
        console.print("from retrieve_virus_data import retrieve_all_virus_data")
        console.print("df = retrieve_all_virus_data()  # Gets all viruses from virus.toml")
        console.print("\n[cyan]The public/private classification will be applied later in processing.[/cyan]")
        
        return 0
        
    except Exception as e:
        console.print(f"[red]❌ Error: {e}[/red]")
        return 1


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)